# Copyright (C) 2013 Intra2net AG
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/lgpl-3.0.html>


import os

from deltatar import crypto
from deltatar.tarfile import TarFile, GNU_FORMAT

import filesplit
from . import BaseTest
from . import new_volume_handler, make_new_encryption_volume_handler

DELTATAR_HEADER_VERSION    = 1
DELTATAR_PARAMETER_VERSION = 1

class EncryptionTest(BaseTest):
    """
    Test encryption after compression in tarfiles
    """

    def test_cli_decrypt(self):
        """
        Create a tar file with only one file inside, using concat
        compression and encryption mode. Then decrypt with crypto.py,
        decompress it with zcat and untar it with gnu tar.

        Note that in an earlier implementation of the Deltatar crypto
        layer, files could be decrypted directly using OpenSSL command
        line tools. With the version 1 parameters this is no longer
        possible since OpenSSL does not ship with a command line tool
        that understands GCM and even if it did, it would be very
        unlikely that it could be made with the raw pdtcrypt format.

        Thus, we rely on the functionality of our encryption library
        ``crypto.py`` to decrypt on the command line; the rest of the
        data pipeline (→ gzip → tar → files) remains the same.
        """

        # create the content of the file to compress and hash it
        hash = self.create_file("big", 50000)

        # create the encryption handler
        encryptor = crypto.Encrypt (password="key",
                                    version=DELTATAR_HEADER_VERSION,
                                    paramversion=DELTATAR_PARAMETER_VERSION)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar.gz.pdtcrypt",
                              mode="w#gz",
                              format=GNU_FORMAT,
                              encryption=encryptor)
        tarobj.add("big")
        tarobj.close()
        os.unlink("big")

        #filesplit.split_file(b'Salted__', "sample.tar.gz.aes.", "sample.tar.gz.aes128")

        # decrypt outer archive layer with crypto.py
        assert os.path.exists("sample.tar.gz.pdtcrypt")
        ret = os.system("python3 -s ./deltatar/crypto.py process key <sample.tar.gz.pdtcrypt >sample.tar.gz")
        assert ret == 0
        assert os.path.exists("sample.tar.gz")

        # extract with normal tar and check output
        os.system("zcat sample.tar.gz 2>/dev/null > sample.tar")
        os.system("tar xf sample.tar")
        assert os.path.exists("big")
        assert hash == self.md5sum("big")

    def test_cli_multiple_files_decrypt(self):
        """
        Create a tar file with multiple files inside, using concat
        compression and encryption mode. Then decrypt and split with
        ``crypto.py``, decompress it with zcat and untar it with gnu tar.
        """

        # create sample data
        hash = dict()
        hash["big"] = self.create_file("big", 50000)
        hash["small"] = self.create_file("small", 100)
        hash["small2"] = self.create_file("small2", 354)

        # create the encryption handler
        encryptor = crypto.Encrypt (password="key",
                                    version=DELTATAR_HEADER_VERSION,
                                    paramversion=DELTATAR_PARAMETER_VERSION)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar.gz.pdtcrypt",
                              mode="w#gz",
                              format=GNU_FORMAT,
                              encryption=encryptor)

        for k in hash:
            tarobj.add(k)
        tarobj.close()

        for k in hash:
            os.unlink(k)

        assert os.path.exists("sample.tar.gz.pdtcrypt")
        ret = os.system("python3 -s ./deltatar/crypto.py process key --split "
                        "-i sample.tar.gz.pdtcrypt -o .")
        assert ret == 0
        for i in range (len (hash)):
            fname = "pdtcrypt-object-%d.bin" % (i + 1)
            assert os.path.exists(fname)
            os.system("zcat '%s' 2>/dev/null > sample.tar" % fname)
            os.system("tar xf sample.tar")

        for key, value in hash.items():
            assert os.path.exists(key)
            assert value == self.md5sum(key)


    def test_decrypt(self):
        """
        Create a tar file with only one file inside, using concat
        compression and encryption mode. Then decrypt it.
        """

        # create the content of the file to compress and hash it
        hash = self.create_file("big", 50000)

        # encryption handling
        encryptor = crypto.Encrypt (password="key",
                                    version=DELTATAR_HEADER_VERSION,
                                    paramversion=DELTATAR_PARAMETER_VERSION)
        decryptor = crypto.Decrypt (password="key")

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar.gz.pdtcrypt",
                              mode="w#gz",
                              format=GNU_FORMAT,
                              encryption=encryptor)
        tarobj.add("big")
        tarobj.close()
        os.unlink("big")

        tarobj = TarFile.open("sample.tar.gz.pdtcrypt",
                              mode="r#gz",
                              format=GNU_FORMAT,
                              encryption = decryptor)
        tarobj.extractall()
        tarobj.close()
        assert os.path.exists("big")
        assert hash == self.md5sum("big")


    def test_multiple_file_decrypt(self):
        """
        Create a tar file with only one file inside, using concat
        compression and encryption mode. Then decrypt it.
        """

        # create sample data
        hash = dict()
        hash["big"] = self.create_file("big", 50000)
        hash["small"] = self.create_file("small", 100)
        hash["small2"] = self.create_file("small2", 354)

        encryptor = crypto.Encrypt (password="key",
                                    version=DELTATAR_HEADER_VERSION,
                                    paramversion=DELTATAR_PARAMETER_VERSION)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar.gz.pdtcrypt",
                              mode="w#gz",
                              format=GNU_FORMAT,
                              encryption=encryptor)

        for k in hash:
            tarobj.add(k)
        tarobj.close()

        for k in hash:
            os.unlink(k)

        decryptor = crypto.Decrypt (password="key")
        tarobj = TarFile.open("sample.tar.gz.pdtcrypt",
                              mode="r#gz",
                              format=GNU_FORMAT,
                              encryption=decryptor)

        tarobj.extractall()
        tarobj.close()

        for key, value in hash.items():
            assert os.path.exists(key)
            assert value == self.md5sum(key)


    def test_multivol_file_decrypt(self):
        '''
        Test multivol tarball with encryption.
        '''

        # create sample data
        hash = dict()
        hash["big"] = self.create_file("big", 50000)
        hash["big2"] = self.create_file("big2", 10200)
        hash["small"] = self.create_file("small", 100)
        hash["small2"] = self.create_file("small2", 354)

        # encryption handler
        encryptor = crypto.Encrypt (password="key",
                                    version=DELTATAR_HEADER_VERSION,
                                    paramversion=DELTATAR_PARAMETER_VERSION)

        # plug the encryption context into the volume handler
        encrypt_volume = make_new_encryption_volume_handler (encryptor)

        # create the tar file with volumes; we need to use a low compression
        # level to force volume split because the test data has too little
        # entropy
        tarobj = TarFile.open("sample.tar.gz.pdtcrypt",
                              mode="w#gz",
                              max_volume_size=20000,
                              compresslevel=0,
                              new_volume_handler=encrypt_volume,
                              encryption=encryptor)

        for k in hash:
            tarobj.add(k)
        tarobj.close()

        assert os.path.exists("sample.tar.gz.pdtcrypt")
        for k in hash:
            os.unlink(k)

        # decrypt
        decryptor = crypto.Decrypt (password="key")

        # plug the encryption context into the volume handler
        decrypt_volume = make_new_encryption_volume_handler (decryptor)

        # extract
        tarobj = TarFile.open("sample.tar.gz.pdtcrypt",
                              mode="r#gz",
                              new_volume_handler=decrypt_volume,
                              encryption=decryptor)
        tarobj.extractall()
        tarobj.close()

        # check output
        for key, value in hash.items():
            assert os.path.exists(key)
            assert value == self.md5sum(key)

